/**
 ******************************************************************************
 * @file           : main.c
 * @author         : Auto-generated by STM32CubeIDE
 * @brief          : Main program body
 ******************************************************************************
 * @attention
 *
 * Copyright (c) 2022 STMicroelectronics.
 * All rights reserved.
 *
 * This software is licensed under terms that can be found in the LICENSE file
 * in the root directory of this software component.
 * If no LICENSE file comes with this software, it is provided AS-IS.
 *
 ******************************************************************************
 */

#include <stdint.h>
#include <stdio.h>

#if !defined(__SOFT_FP__) && defined(__ARM_FP)
  #warning "FPU is not initialized, but the project is compiling for an FPU. Please initialize the FPU before use."
#endif

/* ITM register addresses */
#define ITM_STIMULUS_PORT0    *((volatile uint32_t*) 0xE0000000 )

/* Send a char through ITM */
void ITM_SendChar(uint8_t ch)
{
  // read FIFO status in bit [0]:
  while(!(ITM_STIMULUS_PORT0 & 1));

  //Write to ITM stimulus port0
  ITM_STIMULUS_PORT0 = ch;
}

/* Override low-level _write system call */
int _write(int file, char *ptr, int len)
{
  int DataIdx;

  for (DataIdx = 0; DataIdx < len; DataIdx++)
  {
    ITM_SendChar(*ptr++);
  }
  return len;
}

__attribute__((naked)) void changeStackToPSP(void) {
  // change PSP
  __asm volatile("LDR R0, =_psp_stack");
  __asm volatile("MSR PSP, R0");

  // set SPSEL bit
  __asm volatile("MRS R1, CONTROL");
  __asm volatile("ORR R1, R1, #2");
  __asm volatile("MSR CONTROL, R1");

  // return
  __asm volatile("BX LR");
}

__attribute__ ((naked)) void changeAccessToUnprivileged(void) {
  // set nPRIV bit
  __asm volatile("MRS R0, CONTROL");
  __asm volatile("ORR R0, R0, #1");
  __asm volatile("MSR CONTROL, r0");

  // return
  __asm volatile("BX LR");
}

/* SVC Number Table
 *  0: Add
 *  1: Sub
 *  2: Mul
 *  3: Div
 */

int32_t add(int32_t x, int32_t y) {
  int32_t result;
  // call SVC
  __asm volatile("SVC #0");
  // copy returned value from R0 to variable
  __asm volatile("MOV %0, R0": "=r"(result));
  // return
  return result;
}

int32_t sub(int32_t x, int32_t y) {
  int32_t result;
  __asm volatile("SVC #1");
  __asm volatile("MOV %0, R0": "=r"(result));
  return result;
}

int32_t mul(int32_t x, int32_t y) {
  int32_t result;
  __asm volatile("SVC #2");
  __asm volatile("MOV %0, R0": "=r"(result));
  return result;
}

int32_t div(int32_t x, int32_t y) {
  int32_t result;
  __asm volatile("SVC #3");
  __asm volatile("MOV %0, R0": "=r"(result));
  return result;
}

__attribute__ ((naked)) void SVC_Handler(void) {
  __asm volatile("TST LR, 4"); // check LR to know which stack is used
  __asm volatile("ITE EQ"); // 2 next instructions are conditional
  __asm volatile("MRSEQ R0, MSP"); // save MSP if bit 2 is 0
  __asm volatile("MRSNE R0, PSP"); // save PSP if bit 2 is 1
  __asm volatile("B SVC_Handler_main"); // pass R0 as the argument
}

void SVC_Handler_main(uint32_t* SP) {
  // get the address of the instruction saved in PC
  uint8_t *pInstruction = (uint8_t*)(SP[6]);

  // go back 2 bytes (16-bit opcode)
  pInstruction -= 2;

  // get the opcode, in little endian
  uint8_t svc_num = *pInstruction;

  switch(svc_num) {
    case 0: // add
    {
      int32_t x = SP[0]; // Argument 0 at R0 saved at SP
      int32_t y = SP[1]; // Argument 1 at R1 saved at SP+1
      int32_t z = x + y;
      SP[0] = z;         // Return value saved back to R0 at SP
    }
      break;
    case 1: // sub
      SP[0] = SP[0] - SP[1];
      break;
    case 2: // mul
      SP[0] = SP[0] * SP[1];
      break;
    case 3: // div
      SP[0] = SP[0] / SP[1];
      break;
    default:
      printf("Invalid SVC number: %d!!!\n", svc_num);
      break;
  }
}

int main(void)
{
  changeStackToPSP();
  changeAccessToUnprivileged();

  int32_t x;

  x = add(20, 10);
  printf("20 + 10 = %ld\n", x);

  x = sub(20, 10);
  printf("20 - 10 = %ld\n", x);

  x = mul(20, 10);
  printf("20 * 10 = %ld\n", x);

  x = div(20, 10);
  printf("20 / 10 = %ld\n", x);

  __asm volatile("SVC #6");

  /* Loop forever */
	for(;;);
}
